TUI {
    layout: vertical;
}

NotoControl {
    layout: horizontal;
    background: $boost;
    height: 5;
    margin: 1;
    min-width: 50;
    padding: 1;
    dock: bottom;
}

TextLog {
    height: 1fr;
    border: solid purple;
}

Button {
    width: 16;
}

#mute {
    dock: left;
}

#reset {
    dock: right;
}